// Author: Pete Kirkham
// tiny quad store benchmark.
//
// test merge sort.

#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <stdlib.h>
#include <iostream>
#include <iomanip>

#include "utils/string_utils.h"
#include "core/quad_sort.h"

/*

*/

void dump_quads (unsigned int* data, size_t start, size_t end)
{
    for (size_t quad_index(start); quad_index < end; ++quad_index)
    {
        std::cout  << quad_index << "\t</" << data[quad_index<<2] << ',' << data[(quad_index<<2)|1] << ',';
        std::cout << data[(quad_index<<2)|2] << ',' << data[(quad_index<<2)|3] << "/>" << std::endl;
    }
}

void dump_merge (unsigned int* data, size_t start, size_t end, size_t i, size_t l, size_t r, size_t term)
{
    std::cout << '[';

    for (size_t index(start); index < end; ++index)
    {
        std::cout << (index == i ? '|' : (((index > start) && (data[((index-1)<<2)|term] > data[((index)<<2)|term])) ? '>' : ' '));
        std::cout << std::setfill ('0') << std::setw(4) << data[(index<<2)|term];
    }

    std::cout << (end == i ? '|' : ' ');

    std::cout << ']';

    std::cout << " i=" << i;
    std::cout << " l=" << l;
    std::cout << " r=" << r;

    std::cout << std::endl << ' ';

    for (size_t index(start); index <= end; ++index)
    {
        std::cout << ' ' << ((index == l || index == r) ? "^^^^" : "    ");
    }

    std::cout << std::endl;
}
#ifdef USE_SSE2
inline void copy_quad (unsigned int* source, unsigned int* buffer)
{
    _mm_store_si128((__m128i*)buffer, _mm_load_si128((__m128i*)source));
}
#else
inline void copy_quad (unsigned int* source, unsigned int* buffer)
{
    buffer[0] = source[0];
    buffer[1] = source[1];
    buffer[2] = source[2];
    buffer[3] = source[3];
}
#endif

#ifdef USE_SSE2_IN_MERGE
inline void copy_quad (__m128i* source, __m128i* buffer)
{
    _mm_store_si128(buffer, _mm_load_si128(source));
}

template <unsigned int term> void merge_quads (unsigned int* data, unsigned int* buffer, size_t start, size_t mid, size_t end)
{
    //~ bool watch ((start == 12) && (end == 16));

    __m128i* d128((__m128i*)data);
    __m128i* l(d128+start);
    __m128i* l_max(d128+mid);
    __m128i* r(l_max);
    __m128i* r_max(d128+end);
    __m128i* i((__m128i*)buffer+start);

    //~ if (watch) dump_merge(data, start, end, i, l, r, term);

    __m128i data_l(_mm_load_si128(l));
    __m128i data_r(_mm_load_si128(r));

    // initial scan
    while ((l < l_max) && (r < r_max))
    {
#ifdef USE_SSE2_AVOID_BRANCHES
        data_l = _mm_load_si128(l);
        data_r = _mm_load_si128(r);
        __m128i cmp(_mm_shuffle_epi32(_mm_cmpgt_epi32(data_l, data_r), 0xaa));
        __m128i val(_mm_or_si128(_mm_and_si128(cmp, data_r),_mm_andnot_si128(cmp, data_l)));

        _mm_store_si128(i, val);

        int inc_r(_mm_extract_epi16(cmp, term<<1) & 1);

        r += inc_r;
        l += 1-inc_r;

#else
        __m128i cmp(_mm_cmpgt_epi32(data_l, data_r));


        if (!_mm_extract_epi16(cmp, term<<1))
        {
            _mm_store_si128(i, data_l);
            ++l;
            data_l = _mm_load_si128(l);
        }
        else
        {
           _mm_store_si128(i, data_r);
            ++r;
            data_r = _mm_load_si128(r);
        }
#endif
        ++i;
    }

    while (l < l_max)
    {
        copy_quad(l, i);
        ++l;
        ++i;
    }

    while (r < r_max)
    {
        copy_quad(r, i);
        ++r;
        ++i;
    }
}
#else
template <unsigned int term> void merge_quads (unsigned int* data, unsigned int* buffer, size_t start, size_t mid, size_t end)
{
    unsigned int* l(data+(start<<2));
    unsigned int* l_max(data+(mid<<2));
    unsigned int* r(l_max);
    unsigned int* r_max(data+(end<<2));
    unsigned int* i(buffer+(start<<2));

    unsigned int data_l(*(l+term));
    unsigned int data_r(*(r+term));

    while ((l < l_max) & (r < r_max))
    {
        if (data_l <= data_r)
        {
            copy_quad(l, i);
            l += 4;
            data_l = *(l+term);
        }
        else
        {
            copy_quad(r, i);
            r += 4;
            data_r = *(r+term);
        }

        i += 4;
    }

    while (l < l_max)
    {
        copy_quad(l, i);
        l += 4;
        i += 4;
    }

    while (r < r_max)
    {
        copy_quad(r, i);
        r += 4;
        i += 4;
    }

#ifdef QUAD_SORT_CHECK
    // check
    for (size_t i_check(start + 1); i_check < end; ++i_check)
    {
        if ((buffer[((i_check-1)<<2)|0] == 0) || (buffer[((i_check-1)<<2)|term] > buffer[(i_check<<2)|term]))
        {
            std::cout << "error sorting quads(" << start << ',' << end << ')' << std::endl;
            dump_quads(data, start, end);
            std::cout << std::endl;
            dump_quads(buffer, start, end);
            std::cout << std::endl;
            exit(-1);
        }
    }
#endif
}
#endif

inline size_t min (const size_t& a, const size_t& b) { return a <= b ? a : b; }

template <size_t term> unsigned int* merge_sort_quads_impl (unsigned int* data, unsigned int* buffer, size_t length)
{
    for (size_t step(1); step < length; step <<= 1)
    {
        size_t step_x2(step << 1);

        for (size_t index(0); index + step < length; index += step_x2)
        {
            merge_quads<term>(data, buffer, index, index + step, min(index + step_x2, length));
        }

        unsigned int* temp(buffer);

        buffer = data;
        data = temp;
    }

    return data;
}


unsigned int* merge_sort_quads (unsigned int* data, unsigned int* buffer, size_t length, size_t term)
{
    switch (term)
    {
        case 0:     return merge_sort_quads_impl<0>(data, buffer, length);
        case 1:     return merge_sort_quads_impl<1>(data, buffer, length);
        case 2:     return merge_sort_quads_impl<2>(data, buffer, length);
        case 3:     return merge_sort_quads_impl<3>(data, buffer, length);
        default:    throw illegal_sort_term(term);
    }
}

template <size_t term> unsigned int* radix16_sort_quads_impl (unsigned int* data, unsigned int* buffer, size_t length)
{
    __m128i*    quad_data((__m128i*)data);
    __m128i*    quad_data_end(quad_data+length);
    __m128i*    quad_buff((__m128i*)buffer);
    __m128i*    quad_buff_end(quad_buff+length);

    const size_t i_max(length);

    int nibble_index[0x80];

    // initialise indices
    for (int i(0); i < 0x80; ++i)
    {
        nibble_index[i] = 0;
    }

    // count occurances
    for (__m128i* p_quad(quad_data); p_quad < quad_data_end; ++p_quad)
    {
        __m128i value(_mm_load_si128(p_quad));
        unsigned short lo_word(_mm_extract_epi16(value, (term<<1)));
        unsigned short hi_word(_mm_extract_epi16(value, (term<<1)|1));
        ++nibble_index[((lo_word    )&0xf)     ];
        ++nibble_index[((lo_word>> 4)&0xf)|0x10];
        ++nibble_index[((lo_word>> 8)&0xf)|0x20];
        ++nibble_index[((lo_word>>12)&0xf)|0x30];
        ++nibble_index[((hi_word    )&0xf)|0x40];
        ++nibble_index[((hi_word>> 4)&0xf)|0x50];
        ++nibble_index[((hi_word>> 8)&0xf)|0x60];
        ++nibble_index[((hi_word>>12)&0xf)|0x70];
    }

    // accumulate indices
    for (int i(0); i < 0x80; i += 0x10)
    {
        int index_total(0);

        for (int j(0); j < 0x10; ++j)
        {
            int index_temp(nibble_index[i|j]);
            nibble_index[i|j] = index_total;
            index_total += index_temp;
        }
    }

    // sort on low half word
    for (__m128i* p_quad(quad_data); p_quad < quad_data_end; ++p_quad)
    {
        __m128i         value(_mm_load_si128(p_quad));
        unsigned short  lo_word(_mm_extract_epi16(value, (term<<1)));
        _mm_store_si128(quad_buff + nibble_index[((lo_word    )&0xf)     ]++, value);
    }

    for (__m128i* p_quad(quad_buff); p_quad < quad_buff_end; ++p_quad)
    {
        __m128i         value(_mm_load_si128(p_quad));
        unsigned short  lo_word(_mm_extract_epi16(value, (term<<1)));
        _mm_store_si128(quad_data + nibble_index[((lo_word>> 4)&0xf)|0x10]++, value);
    }

    for (__m128i* p_quad(quad_data); p_quad < quad_data_end; ++p_quad)
    {
        __m128i         value(_mm_load_si128(p_quad));
        unsigned short  lo_word(_mm_extract_epi16(value, (term<<1)));
        _mm_store_si128(quad_buff + nibble_index[((lo_word>> 8)&0xf)|0x20]++, value);
    }

    for (__m128i* p_quad(quad_buff); p_quad < quad_buff_end; ++p_quad)
    {
        __m128i         value(_mm_load_si128(p_quad));
        unsigned short  lo_word(_mm_extract_epi16(value, (term<<1)));
        _mm_store_si128(quad_data + nibble_index[((lo_word>>12)&0xf)|0x30]++, value);
    }

    // sort on high half word
    for (__m128i* p_quad(quad_data); p_quad < quad_data_end; ++p_quad)
    {
        __m128i         value(_mm_load_si128(p_quad));
        unsigned short  hi_word(_mm_extract_epi16(value, (term<<1)|1));
        _mm_store_si128(quad_buff + nibble_index[((hi_word    )&0xf)|0x40]++, value);
    }

    for (__m128i* p_quad(quad_buff); p_quad < quad_buff_end; ++p_quad)
    {
        __m128i         value(_mm_load_si128(p_quad));
        unsigned short  hi_word(_mm_extract_epi16(value, (term<<1)|1));
        _mm_store_si128(quad_data + nibble_index[((hi_word>> 4)&0xf)|0x50]++, value);
    }

    for (__m128i* p_quad(quad_data); p_quad < quad_data_end; ++p_quad)
    {
        __m128i         value(_mm_load_si128(p_quad));
        unsigned short  hi_word(_mm_extract_epi16(value, (term<<1)|1));
        _mm_store_si128(quad_buff + nibble_index[((hi_word>> 8)&0xf)|0x60]++, value);
    }

    for (__m128i* p_quad(quad_buff); p_quad < quad_buff_end; ++p_quad)
    {
        __m128i         value(_mm_load_si128(p_quad));
        unsigned short  hi_word(_mm_extract_epi16(value, (term<<1)|1));
        _mm_store_si128(quad_data + nibble_index[((hi_word>>12)&0xf)|0x70]++, value);
    }

    return data;
}

unsigned int* radix16_sort_quads (unsigned int* data, unsigned int* buffer, size_t length, size_t term)
{
    switch (term)
    {
        case 0:     return radix16_sort_quads_impl<0>(data, buffer, length);
        case 1:     return radix16_sort_quads_impl<1>(data, buffer, length);
        case 2:     return radix16_sort_quads_impl<2>(data, buffer, length);
        case 3:     return radix16_sort_quads_impl<3>(data, buffer, length);
        default:    throw illegal_sort_term(term);
    }
}

// using n implies using n - 1
#ifdef USE_XMM3
#define USE_XMM1
#endif

template <size_t term>
unsigned int* radix256_sort_quads_impl (unsigned int* data, unsigned int* buffer, size_t length)
{
    __m128i*    quad_data((__m128i*)data);
    __m128i*    quad_data_end(quad_data+length);
    __m128i*    quad_buff((__m128i*)buffer);
    __m128i*    quad_buff_end(quad_buff+length);

    const size_t i_max(length);

    int byte0_index[256];
    int byte1_index[256];
    int byte2_index[256];
    int byte3_index[256];

    // initialise indices
    for (int i(0); i < 256; ++i)
    {
        byte0_index[i] = 0;
        byte1_index[i] = 0;
        byte2_index[i] = 0;
        byte3_index[i] = 0;
    }

    // count occurances
    for (__m128i* p_quad(quad_data); p_quad < quad_data_end; ++p_quad)
    {
        __m128i value0(_mm_load_si128(p_quad));
#ifdef USE_XMM1
        __m128i value1(_mm_load_si128(++p_quad));
#ifdef USE_XMM3
        __m128i value2(_mm_load_si128(++p_quad));
        __m128i value3(_mm_load_si128(++p_quad));
#endif
#endif

        unsigned short lo_word0(_mm_extract_epi16(value0, (term<<1)  ));
        ++byte0_index[(lo_word0   )&0xff];
        ++byte1_index[(lo_word0>>8)&0xff];
#ifdef USE_XMM1
        unsigned short lo_word1(_mm_extract_epi16(value1, (term<<1)  ));
        ++byte0_index[(lo_word1   )&0xff];
        ++byte1_index[(lo_word1>>8)&0xff];
#ifdef USE_XMM3
        unsigned short lo_word2(_mm_extract_epi16(value2, (term<<1)  ));
        ++byte0_index[(lo_word2   )&0xff];
        ++byte1_index[(lo_word2>>8)&0xff];

        unsigned short lo_word3(_mm_extract_epi16(value3, (term<<1)  ));
        ++byte0_index[(lo_word3   )&0xff];
        ++byte1_index[(lo_word3>>8)&0xff];
#endif
#endif

        unsigned short hi_word0(_mm_extract_epi16(value0, (term<<1)|1));
        ++byte2_index[(hi_word0   )&0xff];
        ++byte3_index[(hi_word0>>8)&0xff];
#ifdef USE_XMM1
        unsigned short hi_word1(_mm_extract_epi16(value1, (term<<1)|1));
        ++byte2_index[(hi_word1   )&0xff];
        ++byte3_index[(hi_word1>>8)&0xff];
#ifdef USE_XMM3
        unsigned short hi_word2(_mm_extract_epi16(value2, (term<<1)|1));
        ++byte2_index[(hi_word2   )&0xff];
        ++byte3_index[(hi_word2>>8)&0xff];

        unsigned short hi_word3(_mm_extract_epi16(value3, (term<<1)|1));
        ++byte2_index[(hi_word3   )&0xff];
        ++byte3_index[(hi_word3>>8)&0xff];
#endif
#endif
    }

    // accumulate indices
    int byte0_index_total(0);
    int byte1_index_total(0);
    int byte2_index_total(0);
    int byte3_index_total(0);

    for (int i(0); i < 256; ++i)
    {
        int byte0_index_temp(byte0_index[i]); byte0_index[i] = byte0_index_total; byte0_index_total += byte0_index_temp;
        int byte1_index_temp(byte1_index[i]); byte1_index[i] = byte1_index_total; byte1_index_total += byte1_index_temp;
        int byte2_index_temp(byte2_index[i]); byte2_index[i] = byte2_index_total; byte2_index_total += byte2_index_temp;
        int byte3_index_temp(byte3_index[i]); byte3_index[i] = byte3_index_total; byte3_index_total += byte3_index_temp;
    }

    // sort on low half word
    // data -> buffer
    for (__m128i* p_quad(quad_data); p_quad < quad_data_end; ++p_quad)
    {
        __m128i     value0(_mm_load_si128(p_quad));
#ifdef USE_XMM1
        __m128i     value1(_mm_load_si128(++p_quad));
#ifdef USE_XMM3
        __m128i     value2(_mm_load_si128(++p_quad));
        __m128i     value3(_mm_load_si128(++p_quad));
#endif
#endif

        unsigned short  i0(((unsigned short)_mm_extract_epi16(value0, (term<<1)  )   )&0xff);
#ifdef USE_XMM1
        unsigned short  i1(((unsigned short)_mm_extract_epi16(value1, (term<<1)  )   )&0xff);
#ifdef USE_XMM3
        unsigned short  i2(((unsigned short)_mm_extract_epi16(value2, (term<<1)  )   )&0xff);
        unsigned short  i3(((unsigned short)_mm_extract_epi16(value3, (term<<1)  )   )&0xff);
#endif
#endif

        _mm_store_si128(quad_buff + byte0_index[i0]++, value0);
#ifdef USE_XMM1
        _mm_store_si128(quad_buff + byte0_index[i1]++, value1);
#ifdef USE_XMM3
        _mm_store_si128(quad_buff + byte0_index[i2]++, value2);
        _mm_store_si128(quad_buff + byte0_index[i3]++, value3);
#endif
#endif
    }

    // buffer -> data
    for (__m128i* p_quad(quad_buff); p_quad < quad_buff_end; ++p_quad)
    {
        __m128i     value0(_mm_load_si128(p_quad));
#ifdef USE_XMM1
        __m128i     value1(_mm_load_si128(++p_quad));
#ifdef USE_XMM3
        __m128i     value2(_mm_load_si128(++p_quad));
        __m128i     value3(_mm_load_si128(++p_quad));
#endif
#endif
        _mm_store_si128(quad_data + byte1_index[((unsigned short)_mm_extract_epi16(value0, (term<<1)  )>>8)&0xff]++, value0);
#ifdef USE_XMM1
        _mm_store_si128(quad_data + byte1_index[((unsigned short)_mm_extract_epi16(value1, (term<<1)  )>>8)&0xff]++, value1);
#ifdef USE_XMM3
        _mm_store_si128(quad_data + byte1_index[((unsigned short)_mm_extract_epi16(value2, (term<<1)  )>>8)&0xff]++, value2);
        _mm_store_si128(quad_data + byte1_index[((unsigned short)_mm_extract_epi16(value3, (term<<1)  )>>8)&0xff]++, value3);
#endif
#endif
    }

    // sort on high half word
    // data -> buffer
    for (__m128i* p_quad(quad_data); p_quad < quad_data_end; ++p_quad)
    {
        __m128i     value0(_mm_load_si128(p_quad));
#ifdef USE_XMM1
        __m128i     value1(_mm_load_si128(++p_quad));
#ifdef USE_XMM3
        __m128i     value2(_mm_load_si128(++p_quad));
        __m128i     value3(_mm_load_si128(++p_quad));
#endif
#endif
        _mm_store_si128(quad_buff + byte2_index[((unsigned short)_mm_extract_epi16(value0, (term<<1)|1)   )&0xff]++, value0);
#ifdef USE_XMM1
        _mm_store_si128(quad_buff + byte2_index[((unsigned short)_mm_extract_epi16(value1, (term<<1)|1)   )&0xff]++, value1);
#ifdef USE_XMM3
        _mm_store_si128(quad_buff + byte2_index[((unsigned short)_mm_extract_epi16(value2, (term<<1)|1)   )&0xff]++, value2);
        _mm_store_si128(quad_buff + byte2_index[((unsigned short)_mm_extract_epi16(value3, (term<<1)|1)   )&0xff]++, value3);
#endif
#endif
    }

    // buffer -> data
    for (__m128i* p_quad(quad_buff); p_quad < quad_buff_end; ++p_quad)
    {
        __m128i     value0(_mm_load_si128(p_quad));
#ifdef USE_XMM1
        __m128i     value1(_mm_load_si128(++p_quad));
#ifdef USE_XMM3
        __m128i     value2(_mm_load_si128(++p_quad));
        __m128i     value3(_mm_load_si128(++p_quad));
#endif
#endif
        _mm_store_si128(quad_data + byte3_index[((unsigned short)_mm_extract_epi16(value0, (term<<1)|1)>>8)&0xff]++, value0);
#ifdef USE_XMM1
        _mm_store_si128(quad_data + byte3_index[((unsigned short)_mm_extract_epi16(value1, (term<<1)|1)>>8)&0xff]++, value1);
#ifdef USE_XMM3
        _mm_store_si128(quad_data + byte3_index[((unsigned short)_mm_extract_epi16(value2, (term<<1)|1)>>8)&0xff]++, value2);
        _mm_store_si128(quad_data + byte3_index[((unsigned short)_mm_extract_epi16(value3, (term<<1)|1)>>8)&0xff]++, value3);
#endif
#endif
    }

    return data;
}

unsigned int* radix256_sort_quads (unsigned int* data, unsigned int* buffer, size_t length, size_t term)
{
    switch (term)
    {
        case 0: return radix256_sort_quads_impl<0>(data, buffer, length);
        case 1: return radix256_sort_quads_impl<1>(data, buffer, length);
        case 2: return radix256_sort_quads_impl<2>(data, buffer, length);
        case 3: return radix256_sort_quads_impl<3>(data, buffer, length);
        default: throw illegal_sort_term(term);
    }
}

#ifdef USE_SSE2
template <size_t term> unsigned int* radix65536_sort_quads_impl (unsigned int* data, unsigned int* buffer, size_t length)
{
    __m128i*    quad_data((__m128i*)data);
    __m128i*    quad_data_end(quad_data+length);
    __m128i*    quad_buff((__m128i*)buffer);
    __m128i*    quad_buff_end(quad_buff+length);

    const size_t i_max(length);

    int lo_index[65536];
    int hi_index[65536];

    // initialise indices
    for (int i(0); i < 65536; ++i)
    {
        lo_index[i] = 0;
        hi_index[i] = 0;
    }

    // count occurances
    for (__m128i* p_quad(quad_data); p_quad < quad_data_end; ++p_quad)
    {
        __m128i value(_mm_load_si128(p_quad));
        ++lo_index[(unsigned short)_mm_extract_epi16(value, (term<<1)  )];
        ++hi_index[(unsigned short)_mm_extract_epi16(value, (term<<1)|1)];
    }

    // accumulate indices
    int lo_total(0);
    int hi_total(0);

    for (int i(0); i < 65536; ++i) {
        int lo_temp(lo_index[i]);

        lo_index[i] = lo_total;
        lo_total += lo_temp;

        int hi_temp(hi_index[i]);

        hi_index[i] = hi_total;
        hi_total += hi_temp;
    }

    // sort on low half word
    for (__m128i* p_quad(quad_data); p_quad < quad_data_end; ++p_quad)
    {
        __m128i     value(_mm_load_si128(p_quad));
        _mm_store_si128(quad_buff + lo_index[(unsigned short)_mm_extract_epi16(value, (term<<1)  )]++, value);
    }

    // sort on high half word
    for (__m128i* p_quad(quad_buff); p_quad < quad_buff_end; ++p_quad)
    {
        __m128i value(_mm_load_si128(p_quad));
        _mm_store_si128(quad_data + hi_index[(unsigned short)_mm_extract_epi16(value, (term<<1)|1)] ++, value);
    }

    return data;
}

unsigned int* radix65536_sort_quads (unsigned int* data, unsigned int* buffer, size_t length, size_t term)
{
    switch (term)
    {
        case 0: return radix65536_sort_quads_impl<0>(data, buffer, length);
        case 1: return radix65536_sort_quads_impl<1>(data, buffer, length);
        case 2: return radix65536_sort_quads_impl<2>(data, buffer, length);
        case 3: return radix65536_sort_quads_impl<3>(data, buffer, length);
        default: throw illegal_sort_term(term);
    }
}
#else
unsigned int* radix65536_sort_quads (unsigned int* data, unsigned int* buffer, size_t length, size_t term)
{
    const size_t i_max(length<<2);

    int lo_index[65536];
    int hi_index[65536];

    for (int i(0); i < 65536; ++i)
    {
        lo_index[i] = 0;
        hi_index[i] = 0;
    }

    // count occurances
    for (int i(0); i < i_max; i += 4)
    {
        unsigned int value(data[i|term]);
        ++lo_index[value&0xffff];
        ++hi_index[(value>>16)&0xffff];
    }

    // accumulate indices
    int lo_total(0);
    int hi_total(0);

    for (int i(0); i < 65536; ++i) {
        int lo_temp(lo_index[i] << 2);

        lo_index[i] = lo_total;
        lo_total += lo_temp;

        int hi_temp(hi_index[i] << 2);

        hi_index[i] = hi_total;
        hi_total += hi_temp;
    }

    // sort on low half word
    for (int i(0); i < i_max; i += 4)
    {
        unsigned int    value_index((data[i|term])&0xffff);
        int*            p_index(lo_index + value_index);
        int             index(*p_index);
        copy_quad(data + i, buffer + index);
        (*p_index) = index + 4;
    }

    // sort on high half word
    for (int i(0); i < i_max; i += 4)
    {
        unsigned int    value_index(((buffer[i|term])>>16)&0xffff);
        int*            p_index(hi_index + value_index);
        int             index(*p_index);
        copy_quad(buffer + i, data + index);
        (*p_index) = index + 4;
    }
    return data;
}
#endif
template <size_t term>
unsigned int* hybrid_sort_quads_impl (unsigned int* data, unsigned int* buffer, size_t length, size_t block_size)
{
    size_t step(block_size);

    for (size_t index(0); index < length; index += step)
    {
        radix256_sort_quads_impl<term>(data + (index<<2),  buffer + (index<<2), min(step, length - index));
    }

    for (; step < length; step <<= 1)
    {
        size_t step_x2(step << 1);

        for (size_t index(0); index + step < length; index += step_x2)
        {
            merge_quads<term>(data, buffer, index, index + step, min(index + step_x2, length));
        }

        unsigned int* temp(buffer);

        buffer = data;
        data = temp;
    }

    return data;
}

unsigned int* hybrid_sort_quads (unsigned int* data, unsigned int* buffer, size_t length, size_t term, size_t block_size)
{
    switch (term)
    {
        case 0:     return hybrid_sort_quads_impl<0>(data, buffer, length, block_size);
        case 1:     return hybrid_sort_quads_impl<1>(data, buffer, length, block_size);
        case 2:     return hybrid_sort_quads_impl<2>(data, buffer, length, block_size);
        case 3:     return hybrid_sort_quads_impl<3>(data, buffer, length, block_size);
        default:    throw illegal_sort_term(term);
    }
}
